@@ -2,22 +2,30 @@ package DataStore::CAS::FS::DirCodec::Universal;
 use 5.0080001;
 use strict;
 use warnings;
-use Carp;
 use Try::Tiny;
+use Carp;
 use JSON 2.53 ();
+require DataStore::CAS::FS::Dir;
+require DataStore::CAS::FS::DirEnt;
+require DataStore::CAS::FS::InvalidUTF8;
+*decode_utf8= *DataStore::CAS::FS::InvalidUTF8::decode_utf8;
 
 use parent 'DataStore::CAS::FS::DirCodec';
 
-require DataStore::CAS::FS::InvalidUTF8;
-require DataStore::CAS::FS::DirEnt;
-
-our $VERSION= '0.010000';
+our $VERSION= '0.011000';
 
 __PACKAGE__->register_format( universal => __PACKAGE__ );
 
 # ABSTRACT: Codec for saving all arbitrary fields of a DirEnt
 
 
+our $_json_coder;
+sub _build_json_coder {
+	DataStore::CAS::FS::InvalidUTF8->add_json_filter(
+		JSON->new->utf8->canonical->convert_blessed, 1
+	);
+}
+
 sub encode {
 	my ($class, $entry_list, $metadata)= @_;
 	ref($metadata) eq 'HASH' or croak "Metadata must be a hashref"
@@ -25,19 +33,23 @@ sub encode {
 
 	my @entries= sort { $a->{name} cmp $b->{name} }
 		map {
-			my %entry= %{ref $_ eq 'HASH'? $_ : $_->as_hash};
-			defined $entry{name} or croak "Can't serialize nameless directory entry: ".encode_json(\%entry);
-			defined $entry{type} or croak "Can't serialize typeless directory entry: ".encode_json(\%entry);
-			\%entry;
+			my $entry= ref $_ eq 'HASH'? $_ : $_->as_hash;
+			defined $entry->{name} or croak "Can't serialize nameless directory entry: ".JSON::encode_json($entry);
+			defined $entry->{type} or croak "Can't serialize typeless directory entry: ".JSON::encode_json($entry);
+			!defined($_) || (ref $_? ref($_)->can("TO_JSON") : &utf8::is_utf8($_) || !($_ =~ /[\x80-\xFF]/))
+				or croak "Can't serialize $entry->{name}, all attributes must be unicode string, or have TO_JSON: '$_'"
+				for values %$entry;
+			$entry;
 		} @$entry_list;
 
-	my $enc= JSON->new->utf8->canonical->convert_blessed;
-	my $json= $enc->encode($metadata || {});
+	$_json_coder ||= _build_json_coder();
+
+	my $json= $_json_coder->encode($metadata || {});
 	my $ret= "CAS_Dir 09 universal\n"
 		."{\"metadata\":$json,\n"
 		." \"entries\":[\n";
 	for (@entries) {
-		$ret .= $enc->encode($_).",\n"
+		$ret .= $_json_coder->encode($_).",\n"
 	}
 
 	# remove trailing comma
@@ -65,9 +77,9 @@ sub decode {
 		$bytes= <$handle>;
 	}
 
-	my $dec= JSON->new()->utf8->canonical->convert_blessed;
-	DataStore::CAS::FS::InvalidUTF8->add_json_filter($dec, 1);
-	my $data= $dec->decode($bytes);
+	$_json_coder ||= _build_json_coder();
+
+	my $data= $_json_coder->decode($bytes);
 	defined $data->{metadata} && ref($data->{metadata}) eq 'HASH'
 		or croak "Directory data is missing 'metadata'";
 	defined $data->{entries} && ref($data->{entries}) eq 'ARRAY'
@@ -96,7 +108,7 @@ DataStore::CAS::FS::DirCodec::Universal - Codec for saving all arbitrary fields
 
 =head1 VERSION
 
-version 0.010100_01
+version 0.010100_02
 
 =head1 SYNOPSIS
 
